# CoRNy -- A Comic-Ray Neutron processing toolbox for python

This is a project developed by Martin Schrön, Erik Nixdorf, Jannis Jakobi, Mandy Kasner, Carmen Zengerle, Rafael Rosolem,
in order to make Cosmic Ray Neutron Sensing Data Analysis accessible for everyone!
Please ask us about coauthorship if you use our code for a publication.

*Note:* CoRNy is renamed from *CORNish PASDy* -- COsmic-Ray Neutron flavored Processing and Analysis of Sensor Data in pYthon -- because we love cereals, and every corn contributed by the user community adds to universality and flavour. Full generalization of this toolbox also for non-cosmic-ray data is one of the many overachieving future milestones. Then, it may be renamed to PASDy again.

*Important note:* This repository will be renamed to "corny" soon. Be warned to adapt your local copy.

<!--*Note:* This version of PASDy is specific for Cosmic-Ray Neutron Sensing (CRNS). A more general version of PASDy is to be developed in the far future.-->

## Contribution

- Did you find a bug or are you missing a feature? Please report your experience in the [project issues](https://git.ufz.de/CRNS/cornish_pasdy/issues).
- Have you resolved errors or added new features? Please push your contributions back to GitLab such that the whole community can benefit from your work. Contact [Martin](mailto:martin.schroen@ufz.de) to receive write permission.

## Usage

The CoRNy package aims to allow various ways of usage depending on the level of detail and interaction required by individual users. It can be used in the following ways:

- **Mouse click** on the file `instantPASDy.py` processes the sensor data instantly. It requires a file `config.cfg` to be present locally. Use this option if you have only one type of sensor for which the exact same analysis needs to be performed over and over again.
- In the **console**, type

        python instantPASDy.py path/to/your/configfile.cfg

    Use this option if you don't like GUIs or you apply batch scripting. Add the parameter `--nonstop` to avoid the final input prompt at the end of the script.
- **Execute** `app/instantPASDy-GUI.py` where the config file can be selected and ammended directly through the interface. (EXPERIMENTAL)
- Work on your own scripts (e.g., using Jupyter notebooks) and **import** the required snippets from the PASDy package:

        # Add the Corny folder to the recognizable paths
        import sys
        sys.path.insert(0, '/path/to/cornish_pasdy')
        # Import
        from corny import *


## Install

### Python 3.*

Has been tested on:

- Windows
    - [Python.org](https://www.python.org/downloads/) distribution
        - version `3.9.*`
        - version `3.7.*`

Does not seem to work on:

- with Anaconda's python 3.6 on Windows (installation issues with package `osmnx`)


### Use Virtual Environments

The concept of virtual environments protects your other python projects from interfering package versions and dependency clashes.
It is also recommended for CoRNy.

To create a new environment after cloning CoRNy, go to the repository folder and execute:

    python -m venv env

Now you can install packages and mess around without disturbing your other python projects. To start working, activate the session with:

    source env/Scripts/activate

If the execution fails on Windows, it's probably a line-ending error. Edit the file `env/Scripts/activate` and change line endings to `LF` (Unix).

To deactivate the session, just type `deactivate`.

### Package installation

#### Linux

You have three options:

- Install latest versions of the required packages. This may not introduce compatibility issues, but hopefully not:

        pip install -r requirements_1-linux.txt
        pip install -r requirements_2.txt

- Find missing packages that you need to install or update. For this, once install the `pipreq` package and execute:

        pipreqs ./

    The list of required packages is saved to `requirements.txt`.

        pip install -r requirements.txt

    Please remove the file `requirements.txt` (but not the one in `env/`) after you are done, as it is user-specific and should not be part of the repo.

- Use the pre-defined package list deployed by the project in `env/requirements.txt`.
    It probably doesn't contain the latest package versions, but it makes sure that the package versions you install are compatible with CoRNy.

        pip install -r env/requirements.txt

#### Windows

Windows has problems with the automatic installation if the GDAL package. That's why we split the installation into two parts:

1. Download pre-compiled packages ("wheels") of `GDAL`, `pyproj`, `Fiona`, `Shapely`, `rasterio`, and `Cartopy` directly from <https://www.lfd.uci.edu/~gohlke/pythonlibs/>.

2. Install them in the suggested order, either one by one (`pip install filename.whl`), or using:

        pip install -r requirements_1-windows.txt

    Note: version numbers and file names my change in the future, so you might want to adapt the file to your needs.

3. Install the remaining "less problematic" packages using:

        pip install -r requirements_2.txt

    Let us know whether you come over any issues.


### Jupyter Lab

Jupyter notebooks are a great way to play with python, here is a short instruction of installation:

    pip install jupyter
    pip install jupyterlab
    jupyter nbextension enable --py --sys-prefix widgetsnbextension
    jupyter labextension install @jupyterlab/git
    pip install --upgrade jupyterlab-git
    jupyter serverextension enable --py jupyterlab_git

Apply beautiful styles:

1. copy the file `docs/custom.css` into your home folder to `~/.jupyter/custom/`
2. copy the file `docs/pasdy.mplstyle` into your home folder to `~/.matplotlib/stylelib/`
3. Install the `Handlee` font family: <https://fonts.google.com/specimen/Handlee?selection.family=Handlee>

Run Jupyter Lab:

    jupyter lab


## Examples

Example station data:

    ./instantPASDy.py test/input/station.cfg

Example rover data:

    ./instantPASDy.py test/input/rover.cfg
    ./instantPASDy.py test/input/rover-styx.cfg
